import numpy as np

from scipy.optimize import leastsq


class EDGE:
    def __init__(self, contours, min_points=10, shape_cut=0.2,
                 area_cut=10., radius_dev_cut=0.5, connectivity_angle=3.):
        """
        Find, characterize, and quantify the shape of edges (i.e. borders)
        using input contours.
        :param contours: An array containing contour list.
        :param min_points: The number of minimum data points in each edge.
        :param shape_cut: An empirical shape factor cut.
        :param area_cut: An empirical area cut.
        :param radius_dev_cut: An empirical radius deviation cut.
        :param connectivity_angle: An maximum angle to connect
        each separated edge.
        """
        # Set global values.
        self.shape_cut = shape_cut
        self.area_cut = area_cut
        self.radius_dev_cut = radius_dev_cut
        self.connectivity_angle = connectivity_angle

        # Set structure.
        self.edges = []
        for i in range(len(contours)):
            # Remove unclosed contours.
            if contours[i][0][0] != contours[i][-1][0] or \
               contours[i][0][1] != contours[i][-1][1] or \
               len(contours[i]) <= min_points:
                continue

            # All variables are self-explaining except the radius_deviation
            # and the connectivity.
            # The radius_deviation is ratio of the standard deviation
            # of the distances from the center to the radius.
            # The connectivity indicates an index of an edge likely to be
            # connected with the current edge. -1 indicates no connectivity.
            self.edges.append({
                'index': i + 1,
                # Note that the contours returned from the scikit-image
                # is a list of [row, columns]
                'x': contours[i][::, 1],
                'y': contours[i][::, 0],
                'x_center': 0., 'y_center': 0.,
                'perimeter': 0., 'area': 0.,
                'shape_factor': 0.,
                'radius_deviation': 0.,
                'slope': 0., 'intercept': 0.,
                'connectivity': -1,
                # For plotting a box surrounding the edge.
                'x_min':0., 'x_max':0.,
                'y_min':0., 'y_max':0.,
                # For plotting. To check if this edge
                # is surrounded by a box already.
                'box_plotted': False})

    def quantify(self):
        """
        Quantify shape of the contours.
        """
        four_pi = 4. * np.pi
        for edge in self.edges:
            # Positions
            x = edge['x']
            y = edge['y']

            A, perimeter, x_center, y_center, distances = \
                self.get_shape_factor(x, y)

            # Set values.
            edge['area'] = A
            edge['perimeter'] = perimeter
            edge['x_center'] = x_center
            edge['y_center'] = y_center
            # Circle is 1. Rectangle is 0.78. Thread-like is close to zero.
            edge['shape_factor'] = four_pi * edge['area'] / \
                                   edge['perimeter'] ** 2.

            # We assume that the radius of the edge
            # as the median value of the distances from the center.
            radius = np.median(distances)
            edge['radius_deviation'] = np.std(distances - radius) / radius

            edge['x_min'] = np.min(x)
            edge['x_max'] = np.max(x)
            edge['y_min'] = np.min(y)
            edge['y_max'] = np.max(y)

    def get_shape_factor(self, x, y):
        """
        Return values related to the shape based on x and y.
        :param x: An array of x coordinates.
        :param y: An array of y coordinates.
        :return: Area, perimeter, x_center, y_center, distances from the center.
        """

        # Area.
        xyxy = (x[:-1] * y[1:] - x[1:] * y[:-1])
        A = 1. / 2. * np.sum(xyxy)

        # X and Y center.
        one_sixth_a = 1. / (6. * A)
        x_center = one_sixth_a * np.sum((x[:-1] + x[1:]) * xyxy)
        y_center = one_sixth_a * np.sum((y[:-1] + y[1:]) * xyxy)

        # Perimeter.
        perimeter = np.sum(np.sqrt((x[1:] - x[:-1])**2 +
                                   (y[1:] - y[:-1])**2))

        # Distances from the center.
        distances = np.sqrt((x - x_center)**2 + (y - y_center)**2)

        return np.abs(A), perimeter, x_center, y_center, distances

    def get_edges(self):
        return self.edges

    def filter_edges(self):
        """
        Remove edges unlikely to be streaks.
        """
        filtered_edges = []
        filtered_cnt = 1
        for edge in self.edges:
            if edge['shape_factor'] <= self.shape_cut and \
               edge['area'] >= self.area_cut and \
               edge['radius_deviation'] >= self.radius_dev_cut:
                # Reset index, incremental from 1.
                edge['index'] = filtered_cnt
                filtered_edges.append(edge)
                filtered_cnt += 1

        # Set filtered edges.
        self.edges = filtered_edges

    def residuals(self, theta, x, y):
        """
        Residual for a straight line.
        :param theta: Coefficients.
        :param x: An array of x values.
        :param y: An array of y values.
        :return: residuals.
        """

        residu = y - (theta[0] * x + theta[1])

        return residu

    def connect_edges(self):
        """
        Connect detected edges based on their slopes.
        """
        # Fitting a straight line to each edge.
        p0 = [0., 0.]
        radian2angle = 180. / np.pi
        for edge in self.edges:
            p1, s = leastsq(self.residuals, p0,
                            args=(edge['x'][:-1],
                                  edge['y'][:-1]))
            edge['slope'] = p1[0]
            edge['intercept'] = p1[1]
            edge['slope_angle'] = np.arctan(edge['slope']) * radian2angle

        # Connect by the slopes of two edges.
        len_edges = len(self.edges)
        for i in range(len_edges - 1):
            for j in range(i + 1, len_edges):
                if np.abs(self.edges[i]['slope_angle'] -
                          self.edges[j]['slope_angle']) <= \
                   self.connectivity_angle:
                    # Then, the slope between the centers of the two edges
                    # should be similar with the slopes of
                    # the two lines of the edges as well.
                    c_slope = (self.edges[i]['y_center'] -
                                    self.edges[j]['y_center']) / \
                                   (self.edges[i]['x_center'] -
                                    self.edges[j]['x_center'])
                    c_slope_angle = np.arctan(c_slope) * radian2angle

                    if np.abs(c_slope_angle - self.edges[i]['slope_angle']) <= \
                       self.connectivity_angle and \
                       np.abs(c_slope_angle - self.edges[j]['slope_angle']) <= \
                       self.connectivity_angle:
                        self.edges[i]['connectivity'] = self.edges[j]['index']
                        break

if __name__ == '__main__':
    import pylab as pl

    # Sample contour with only one edge.
    contours = np.array([[[ 985.2156529 ,  385.        ],
                         [ 985.21551898,  384.        ],
                         [ 985.        ,  383.78486061],
                         [ 984.        ,  383.7864848 ],
                         [ 983.21207242,  383.        ],
                         [ 983.21186237,  382.        ],
                         [ 983.        ,  381.83091953],
                         [ 982.        ,  381.16390331],
                         [ 981.93404467,  381.        ],
                         [ 982.        ,  380.95885879],
                         [ 982.96034405,  380.        ],
                         [ 982.98009999,  379.        ],
                         [ 983.        ,  378.96023005],
                         [ 983.64027253,  378.        ],
                         [ 983.91887289,  377.        ],
                         [ 983.83748505,  376.        ],
                         [ 983.        ,  375.16429886],
                         [ 982.8353446 ,  375.        ],
                         [ 982.83521735,  374.        ],
                         [ 983.        ,  373.33646663],
                         [ 983.08356296,  373.        ],
                         [ 983.08346613,  372.        ],
                         [ 983.05559534,  371.        ],
                         [ 983.        ,  370.91681365],
                         [ 982.08104495,  370.        ],
                         [ 982.        ,  369.34829976],
                         [ 981.8839432 ,  369.        ],
                         [ 981.88365281,  368.        ],
                         [ 981.        ,  367.67102221],
                         [ 980.67173695,  368.        ],
                         [ 980.67158503,  369.        ],
                         [ 981.        ,  369.52649409],
                         [ 981.78545036,  370.        ],
                         [ 982.        ,  370.0808727 ],
                         [ 982.91699005,  371.        ],
                         [ 982.        ,  371.91889354],
                         [ 981.94108342,  372.        ],
                         [ 981.7843259 ,  373.        ],
                         [ 981.        ,  373.16891089],
                         [ 980.70386539,  373.        ],
                         [ 980.        ,  372.29777757],
                         [ 979.        ,  372.26617816],
                         [ 978.19968936,  372.        ],
                         [ 978.        ,  371.60280545],
                         [ 977.        ,  371.80292069],
                         [ 976.        ,  371.80503356],
                         [ 975.61260341,  372.        ],
                         [ 976.        ,  372.19512255],
                         [ 976.4018599 ,  373.        ],
                         [ 976.        ,  373.22993289],
                         [ 975.        ,  373.61366516],
                         [ 974.        ,  373.61765716],
                         [ 973.44589678,  374.        ],
                         [ 973.11426113,  375.        ],
                         [ 973.03984692,  376.        ],
                         [ 973.05362787,  377.        ],
                         [ 973.        ,  377.4890333 ],
                         [ 972.        ,  377.49473241],
                         [ 971.68930353,  377.        ],
                         [ 971.        ,  376.82882956],
                         [ 970.65813729,  377.        ],
                         [ 971.        ,  377.49982635],
                         [ 971.68492286,  378.        ],
                         [ 972.        ,  378.1577691 ],
                         [ 973.        ,  378.15951288],
                         [ 973.38907309,  379.        ],
                         [ 973.72426519,  380.        ],
                         [ 973.        ,  380.83857605],
                         [ 972.83871315,  381.        ],
                         [ 973.        ,  381.65355224],
                         [ 974.        ,  381.64301045],
                         [ 975.        ,  381.64618349],
                         [ 975.17591572,  382.        ],
                         [ 976.        ,  382.82588456],
                         [ 977.        ,  382.82742928],
                         [ 977.22970369,  383.        ],
                         [ 977.07642179,  384.        ],
                         [ 977.        ,  384.08592729],
                         [ 976.08630625,  385.        ],
                         [ 976.        ,  385.19764976],
                         [ 975.29973193,  386.        ],
                         [ 976.        ,  386.7016911 ],
                         [ 976.29771644,  387.        ],
                         [ 977.        ,  387.46943243],
                         [ 977.52989259,  388.        ],
                         [ 978.        ,  388.70553523],
                         [ 979.        ,  388.70677412],
                         [ 979.58307735,  389.        ],
                         [ 980.        ,  389.27882521],
                         [ 980.54135193,  390.        ],
                         [ 981.        ,  390.45941675],
                         [ 981.45976149,  390.        ],
                         [ 982.        ,  389.69196352],
                         [ 982.6064398 ,  389.        ],
                         [ 982.        ,  388.3940689 ],
                         [ 981.2106587 ,  388.        ],
                         [ 982.        ,  387.21142985],
                         [ 983.        ,  387.2128304 ],
                         [ 984.        ,  387.47572913],
                         [ 984.3566691 ,  387.        ],
                         [ 984.        ,  386.28734439],
                         [ 983.42505279,  386.        ],
                         [ 984.        ,  385.42651258],
                         [ 985.        ,  385.42957929],
                         [ 985.2156529 ,  385.        ]]])

    edge = EDGE(contours)
    edge.quantify()
    edge.connect_edges()
    edges = edge.get_edges()
    print edges

    # Plot the first edge, which is the only edge in the sample data.
    # Since the contour returned from scikit-image is a list of
    # [row, columns], we need to use "1" as x and "0" as y.
    pl.plot(contours[0][::, 1], contours[0][::, 0], 'r+-')
    window_size = 10
    x_added = np.hstack([contours[0][::, 1], contours[0][::, 1][:window_size]])
    y_added = np.hstack([contours[0][::, 0], contours[0][::, 0][:window_size]])
    pl.plot(edges[0]['x_center'], edges[0]['y_center'], 'rs')

    # Plot fitting line.
    x = np.linspace(365, 395, 10)
    y = edges[0]['slope'] * x + edges[0]['intercept']
    pl.plot(x, y, 'b-')
    pl.show()



__author__ = 'kim'

"""
Base IO code for all datasets
"""

from os.path import dirname
from os.path import join

from astropy.io import fits


def read_fits(filename='long.fits'):
    """
    Read an sample fits file and return only image part.

    :param filename: Fits filename.
    :return: Image data.
    """

    module_path = dirname(__file__)
    file_path = join(module_path, 'samples', filename)

    hdulist = fits.open(file_path)

    data = hdulist[0].data

    hdulist.close()

    return data


import os

import numpy as np
import pylab as pl

from skimage import measure
from astropy.io import fits
#from astropy.stats import sigma_clipped_stats
from photutils.background import Background

from astride.utils.edge import EDGE


class Streak:
    def __init__(self, filename, bkg_box_size=50, contour_threshold=3.,
                 min_points=10, shape_cut=0.2, area_cut=10.,
                 radius_dev_cut=0.5, connectivity_angle=3.,
                 output_path=None):
        """
        Initialize the streak instance.
        :param filename: Fits filename.
        :param bkg_box_size: Box size for background estimation.
        :param contour_threshold: Threshold to search contours (i.e. edges of
        an input image)
        :param min_points: The number of minimum data points in each edge.
        :param shape_cut: An empirical shape factor cut.
        :param area_cut: An empirical area cut.
        :param radius_dev_cut: An empirical radius deviation cut.
        :param connectivity_angle: An maximum angle to connect
        each separated edge.
        :param output_path: Path to save figures and output files. If None,
        the base filename is used as the folder name.
        """
        hdulist = fits.open(filename)
        raw_image = hdulist[0].data.astype(np.float64)
        hdulist.close()

        # Raw image.
        self.raw_image = raw_image
        # Background structure and background map
        self.__bkg = None
        self.background_map = None
        # Background removed image.
        self.image = None
        # Raw edges
        self.raw_edges = None
        # Filtered edges, so streak, by their morphologies and
        # also connected (i.e. linked) by their slope.
        self.streaks = None

        # Other variables.
        self.bkg_box_size = bkg_box_size
        self.contour_threshold = contour_threshold

        # These variables for the edge detections and linking.
        self.min_points = min_points
        self.shape_cut = shape_cut
        self.area_cut = area_cut
        self.radius_dev_cut = radius_dev_cut
        self.connectivity_angle = connectivity_angle

        # Set output path.
        if output_path is None:
            output_path = './%s/' % (os.path.basename(filename).split('.')[0])
        if output_path[-1] != '/':
            output_path += '/'
        self.output_path = output_path

        # For plotting.
        pl.rcParams['figure.figsize'] = [12, 9]

    def detect(self):
        """
        Run the pipeline to detect streaks.
        """
        self.__remove_background()
        #self.__detect_sources()
        self.__detect_streaks()

    def __remove_background(self):
        # Get background map and subtract.
        self.__bkg = Background(self.raw_image,
                                (self.bkg_box_size, self.bkg_box_size),
                                method='median')
        self.background_map = self.__bkg.background
        self.image = self.raw_image - self.background_map

    def __detect_streaks(self):
        # Find contours.
        # Returned contours is the list of [row, columns] (i.e. [y, x])
        bkg_rms = self.__bkg.background_rms_median
        contours = measure.find_contours(
            self.image, bkg_rms * self.contour_threshold, fully_connected='high'
                                         )

        # Quantify shapes of the contours and save them as 'edges'.
        edge = EDGE(contours, min_points=self.min_points,
                    shape_cut=self.shape_cut, area_cut=self.area_cut,
                    radius_dev_cut=self.radius_dev_cut,
                    connectivity_angle=self.connectivity_angle)
        edge.quantify()
        self.raw_edges = edge.get_edges()

        # Filter the edges, so only streak remains.
        edge.filter_edges()
        edge.connect_edges()
        self.streaks = edge.get_edges()

    def __detect_sources(self):
        from photutils import daofind

        fwhm = 3.
        detection_threshold = 3.
        sources = daofind(self.image,
                          threshold=self.__bkg.background_rms_median *
                                    detection_threshold,
                          fwhm=fwhm)
        #sources = irafstarfind(data, threshold=std * detection_threshold, fwhm=FWHM)
        pl.plot(sources['xcentroid'], sources['ycentroid'], 'r.')

    def __find_box(self, n, edges, xs, ys):
        """
        Recursive function that defines a box surrounding
        one or more edges that are connected to each other.
        :param n: Index of edge currently checking.
        :param edges: edges.
        :param xs: x min and max coordinates.
        :param ys: y min and max coordinates.
        :return: x and y coordinates for plot plotting.
        """
        # Add current coordinates.
        current_edge = [edge for edge in edges if edge['index'] == n][0]
        current_edge['box_plotted'] = True
        xs.append([current_edge['x_min'], current_edge['x_max']])
        ys.append([current_edge['y_min'], current_edge['y_max']])

        # If connected with other edge.
        if current_edge['connectivity'] != -1:
            self.__find_box(current_edge['connectivity'], edges, xs, ys)
        # Otherwise.
        else:
            return xs, ys

    def plot_figures(self, cut_threshold=5.):
        """
        Save figures of detected streaks under the 'path' folder.
        :param cut_threshold: Threshold to cut image values to make it
        more visible.
        """
        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)

        # Plot the image.
        plot_data = self.image.copy()
        #mean, med, std = sigma_clipped_stats(self.image, sigma=3.0, iters=5)
        # Background subtracted image, so the median value should be close to zero.
        med = 0.
        std = self.__bkg.background_rms_median
        plot_data[np.where(self.image > med + cut_threshold * std)] = \
            med + cut_threshold * std
        plot_data[np.where(self.image < med - cut_threshold * std)] = \
            med - cut_threshold * std
        pl.imshow(plot_data, origin='lower', cmap='gray')

        edges = self.streaks
        # Plot all contours.
        for n, edge in enumerate(edges):
            pl.plot(edge['x'], edge['y'])
            pl.text(edge['x'][0], edge['y'][1],
                    '%d' % (edge['index']), color='b', fontsize=15)

        # Plot boxes.
        # Box margin in pixel.
        box_margin = 10
        for n, edge in enumerate(edges):
            # plot boxes around the edge.
            if not edge['box_plotted']:
                # Define the box to plot.
                xs = []
                ys = []
                self.__find_box(edge['index'], edges, xs, ys)
                x_min = max(np.min(xs) - box_margin, 0)
                x_max = min(np.max(xs) + box_margin, self.image.shape[0])
                y_min = max(np.min(ys) - box_margin, 0)
                y_max = min(np.max(ys) + box_margin, self.image.shape[1])
                box_x = [x_min, x_min, x_max, x_max]
                box_y = [y_min, y_max, y_max, y_min]
                pl.fill(box_x, box_y, ls='--', fill=False, ec='r', lw=2)
                edge['box_plotted'] = True

        pl.xlabel('X/pixel')
        pl.ylabel('Y/pixel')
        pl.axis([0, self.image.shape[0], 0, self.image.shape[1]])
        pl.savefig('%sall.png' % (self.output_path))

        # Plot all individual edges (connected).
        for n, edge in enumerate(edges):
            # Reset.
            edge['box_plotted'] = False

        for n, edge in enumerate(edges):
            if not edge['box_plotted']:
                # Define the box to plot.
                xs = []
                ys = []
                self.__find_box(edge['index'], edges, xs, ys)
                x_min = max(np.min(xs) - box_margin, 0)
                x_max = min(np.max(xs) + box_margin, self.image.shape[0])
                y_min = max(np.min(ys) - box_margin, 0)
                y_max = min(np.max(ys) + box_margin, self.image.shape[1])
                edge['box_plotted'] = True
                pl.axis([x_min, x_max, y_min, y_max])
                pl.savefig('%s%d.png' % (self.output_path, edge['index']))

        # Clear figure.
        pl.clf()

    def write_outputs(self):
        """
        Write information of detected streaks under the 'path' folder.
        :param path:
        """

        if not os.path.exists(self.output_path):
            os.makedirs(self.output_path)

        fp = open('%sstreaks.txt' % (self.output_path), 'w')
        fp.writelines('#ID x_center y_center area perimeter shape_factor ' +
                      'radius_deviation slope_angle intercept connectivity\n')
        for n, edge in enumerate(self.streaks):
            line = '%2d %7.2f %7.2f %6.1f %6.1f %6.3f %6.2f %5.2f %7.2f %2d\n' % \
                   (
                       edge['index'], edge['x_center'], edge['y_center'],
                       edge['area'], edge['perimeter'], edge['shape_factor'],
                       edge['radius_deviation'], edge['slope_angle'],
                       edge['intercept'], edge['connectivity']
                   )
            fp.writelines(line)
        fp.close()

if __name__ == '__main__':
    import time

    #streak = Streak('/Users/kim/Dropbox/python/ASTRiDE/astride/datasets/samples/long.fits')
    streak = Streak('./datasets/samples/long.fits')

    start = time.time()
    streak.detect()
    end = time.time()

    print streak.streaks

    streak.plot_figures()
    streak.write_outputs()

    print '%.2f seconds' % (end - start)



__author__ = 'kim'

import logging


class Logger():
    """
    Create logger instance for writing to both console and a local file.
    """
    def __init__(self, filepath=None):
        """
        Initailize. Write logger to either console and/or to a disk.

        :param filepath: specify a log filename with the absolute path.
            If not given, no output is written to a file.
        :return:
        """

        # create logger.
        logger = logging.getLogger('UPSILoN')
        logger.setLevel(logging.DEBUG)

        # create formatter and add it to the handlers
        formatter = logging.Formatter('%(asctime)s %(levelname)s - %(message)s')
            #datefmt='%Y/%m/%d %H:%M:%S')

        # create file handler which logs even debug messages.
        if filepath and filepath[0] == '/':
            fh = logging.FileHandler(filepath, 'w')
            fh.setLevel(logging.DEBUG)
            fh.setFormatter(formatter)
            logger.addHandler(fh)

        # create console handler with a higher log level.
        ch = logging.StreamHandler()
        ch.setLevel(logging.INFO)
        ch.setFormatter(formatter)

        # add the handlers to logger
        logger.addHandler(ch)

        self.logger = logger

    def getLogger(self):
        """
        Return log instance.
        """

        return self.logger

if __name__ == '__main__':

    logger = Logger().getLogger()

    logger.debug('debug message')
    logger.info('info message')
    logger.warn('warn message')
    logger.error('error message')
    logger.critical('critical message')


from os.path import dirname
from os.path import join

from astride.detect import Streak
from astride.utils.logger import Logger


def test():
    logger = Logger().getLogger()

    logger.info('Start.')
    module_path = dirname(__file__)
    file_path = join(module_path, '../datasets/samples', 'long.fits')

    logger.info('Read a fits file..')
    streak = Streak(file_path)

    logger.info('Search streaks..')
    streak.detect()

    logger.info('Save figures and write outputs to %s' %
                streak.output_path)
    streak.plot_figures()
    streak.write_outputs()

    logger.info('Done.')

    logger.handlers = []


if __name__ == '__main__':
    test()


from setuptools import find_packages, setup

def readme():
    with open('README.rst') as f:
        return f.read()

setup(
    name='astride',
    version='0.2',
    description='Automated Streak Detection for High Velocity Objects',
    long_description=readme(),
    platforms=['any'],
    packages=find_packages(),
    include_package_data=True,
    url='https://github.com/dwkim78/astride',
    license='MIT',
    author='Dae-Won Kim',
    author_email='dwkim78@gmail.com',
    install_requires=['numpy>=1.10.4', 'photutils>=0.2.1', 'astropy>=1.1.1',
                      'matplotlib>=1.5.1', 'scipy>=0.16.1',
                      'scikit-image>=0.11.3'],
    keywords=['astronomy', 'image', 'streak', 'satellite', 'meteor'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Astronomy'
    ]
)



import numpy as np


def moving_average(data, window_size=5):
    """

    :param data: An array of values.
    :param window_size: Moving average window size.
    :return: Moving averaged array.
    """
    window = np.ones(int(window_size))/float(window_size)
    results = np.convolve(data, window, 'valid')

    return results